import argparse
import sys

# TODO Figure out why I need different imports for pip and for testing
try:
    from .parser import parse_command
    from .environment import *
except ImportError:
    from parser import parse_command
    from environment import *


def main() -> None:
    parser = argparse.ArgumentParser(description="Python with pipes")
    parser.add_argument("command", help="the pypthon command")
    parser.add_argument("-s", "--show-python", help="show python code generated", action="store_true")
    parser.add_argument("-i", "--import", help="module to import eg: -i 'requests' -i 'numpy'", default=[], action="append")
    parser.add_argument("-d", "--dry-run", help="don't run python code", action="store_true")
    parser.add_argument("-e", "--environment", help="show variables excluding -i imports in environment", action="store_true")
    args = parser.parse_args()
    python = parse_command(args.command)
    if args.show_python or args.dry_run:
        print(python)
    code = '\n'.join([
        get_user_custom_setup(),
        python
    ])
    for module in getattr(args, "import"):
        import_string = f"import {module}"
        if not args.dry_run:
            exec(import_string, globals())
    if args.dry_run or args.environment:
        pprint(sorted(v for v in globals() if v[:1] != "_"))
    if not args.dry_run:
        exec(code)


def get_user_custom_setup() -> str:
    pyprc = Path.home() / ".pypthonrc.py"
    pyprc.touch()
    with open(pyprc, 'r') as fin:
        return fin.read()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



from typing import Generator, List, Tuple

literal_chars = {'"': '"', "'": "'", "(": ")", "[": "]", "{": "}"}
operator_chars = {
    "*",
    "**",
    "^",
    "&",
    "*",
    "<",
    "<=",
    "==",
    ">",
    ">=",
    "!=",
    "+",
    "-",
    "/",
    "//",
    "<<",
    ">>",
    ",",
}
spaced_operator_chars = {
    "and",
    "or",
    "in",
    "not in",
    "if",
    "else",
}


def parse_command(command: str) -> str:
    """
    Turns a command into valid python.
    """
    segments = get_piped_segments(command)
    value = segments[0]

    if 1 or __is_literal(value):
        segments = segments[1:]
    else:
        value = "stdin"

    for segment in segments:
        value = __combine_two_piped(value, segment)

    return value


def get_piped_segments(command: str) -> List[str]:
    """
    Gets the python code of each command in between pipes.
    """
    return [
        pypthon_cmd_to_python(term, i == 0)
        for i, term in enumerate(__term_enumerator(command, "|"))
    ]


def pypthon_cmd_to_python(command: str, is_first: bool) -> str:
    command = command.strip()

    if __is_literal(command):
        return command

    name, args = __get_function_name_and_args(command)

    if is_first:
        return name
    return f"{name}({', '.join(args)})"


def __is_literal(code: str) -> bool:
    code = code.strip()
    return literal_chars.get(code[:1]) == code[-1:]


def __get_function_name_and_args(command: str) -> Tuple[str, List[str]]:
    command = command.strip()
    try:
        sidx = command.index(" ")
        return command[:sidx], __spaces_separated_to_args(command[sidx:])
    except ValueError:
        return command, []


def __spaces_separated_to_args(space_arg: str) -> List[str]:
    terms = list(__term_enumerator(space_arg, " "))
    return __combine_args_with_operators(terms)


def __term_enumerator(terms: str, separator: str) -> Generator[str, None, None]:
    last_idx = 0
    is_creating_literal = []
    ignore = {"not"}

    for i, c in enumerate(terms + separator):
        if is_creating_literal and c == literal_chars.get(is_creating_literal[-1]):
            is_creating_literal.pop()
        elif c in literal_chars:
            is_creating_literal.append(c)
        elif c == separator and not is_creating_literal:
            term = terms[last_idx:i].strip()
            if term and term not in ignore:
                yield term
                last_idx = i + 1


def __combine_args_with_operators(args: List[str]) -> List[str]:
    args = args[:]
    args_after_operators = []

    while args:
        arg = args.pop()
        is_colon = arg[-1:] == ":"
        if (
            arg in spaced_operator_chars
            or arg[-1:] in operator_chars
            or arg[-2:] in operator_chars
            or is_colon
        ):
            prev = args_after_operators[-1]
            prev_args = __get_prev_args(args, arg)
            if is_colon:
                args_after_operators[-1] = f"lambda {prev_args}{arg} {prev}"
            else:
                other_value = args.pop()
                args_after_operators[-1] = f"{other_value} {arg} {prev}"
        else:
            args_after_operators.append(arg)

    return args_after_operators[::-1]


def __get_prev_args(args: List[str], initial: str) -> str:
    prev_args = []
    repeat = initial.strip().startswith(",")
    while args:
        if args[-1].strip().endswith(",") or repeat:
            arg = args.pop().strip()
            repeat = arg.startswith(",")
            prev_args.append(arg)
        else:
            break
    if prev_args:
        return " ".join(prev_args) + " "
    return ""


def __combine_two_piped(first: str, second: str) -> str:
    if second[-2:-1] != "(":
        return f"{second[:-1]}, {first}{second[-1:]}"
    return f"{second[:-1]}{first}{second[-1:]}"



import functools
import itertools as it
import os
import random
import re
import shlex
import shutil
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from functools import reduce
from pathlib import Path
from pprint import pprint
from sys import stdin


__executor = ThreadPoolExecutor()


def iterable_first(fn):
    """
    Decorator used to make the piped value the first argument.
    """
    @functools.wraps(fn)
    def inner(*args):
        *regular_args, piped = args
        return fn(piped, *regular_args)
    return inner


@iterable_first
def uprint(iterable, kwargs={}):
    """
    Allows unpacking into print.

    Usage:

    >>> ["hello", "world"] | uprint {"sep": ":"}
    hello:world
    >>> ["hello", "world"] | uprint
    hello world
    """
    return print(*iterable, **kwargs)
    

@iterable_first
def ufunc(iterable, func, kwargs={}):
    """
    Allows unpacking iterable into functions.

    Usage:

    >>> ["hello", "world"] | ufunc print {"sep": ":"}
    hello:world
    >>> ["hello", "world"] | ufunc print
    hello world
    """
    return func(*iterable, **kwargs)


def sh(command, iterable):
    """
    Opens a shell function.

    This streams both input and the response.
    """
    process = __open_process(command)
    __executor.submit(__write_to_process, process, iterable)
    return process.stdout


@iterable_first
def cmap(iterable, func, threads=100):
    """
    Concurrent map.

    Usage:

    >>>  range(3) | cmap requests.get("https://google.com") | uprint
    <Response [200]> <Response [200]> <Response [200]>
    """
    executor = ThreadPoolExecutor(max_workers=threads)
    return executor.map(func, iterable)


@iterable_first
def cfilter(iterable, func, threads=100):
    """
    Concurrent filter.

    Usage:

    >>> ['https://google.com', 'https://google.com/pypthon'] | cfilter url: requests.get(url).status_code == 200 | uprint
    https://google.com
    """
    def map_fn(value):
        return func(value), value

    for is_valid, value in cmap(map_fn, threads, iterable):
        if is_valid:
            yield value


def result_fn(func):
    """
    Wrapper to return either the function return or the exception raised.
    """
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return e
    return inner


def __open_process(command):
    pipe = subprocess.PIPE
    return subprocess.Popen(
        shlex.split(command),
        universal_newlines=True,
        stdin=pipe,
        stdout=pipe,
        stderr=pipe,
    )


def __write_to_process(process, iterable):
    for line in iterable:
        process.stdin.write(line + "\n")
        process.stdin.flush()
    process.stdin.close()



from .context import pypthon

parser = pypthon.parser


def test_parse_command() -> None:
    test_cases = [
        ['"hello" | print', 'print("hello")'],
        [
            "[1, 2, 3, 4] | map x: x + 1 | list | print",
            "print(list(map(lambda x: x + 1, [1, 2, 3, 4])))",
        ]
        # ["map x: x + 1 | list | print", "print(list(map(lambda x: x + 1, stdin)))"],
    ]
    for case in test_cases:
        assert parser.parse_command(*case[:-1]) == case[-1]


def test_get_piped_segments() -> None:
    test_cases = [
        ['"hello" | print', ['"hello"', "print()"]],
        [
            "[1, 2, 3, 4] | map x: x + 1 | list | print",
            ["[1, 2, 3, 4]", "map(lambda x: x + 1)", "list()", "print()"],
        ],
        [
            '"Hello with a | in between" | print',
            ['"Hello with a | in between"', "print()"],
        ],
    ]

    for case in test_cases:
        assert parser.get_piped_segments(*case[:-1]) == case[-1]


def test_pypthon_cmd_to_python() -> None:
    test_cases = [
        ["print", "print()"],
        ["print 'hello' 'world'", "print('hello', 'world')"],
        ["print 1 + 2", "print(1 + 2)"],
        ["map x: x + 1", "map(lambda x: x + 1)"],
        ["map x: x[1: 2]", "map(lambda x: x[1: 2])"],
        ["filter x: x[: 2] != 'oi'", "filter(lambda x: x[: 2] != 'oi')"],
    ]

    for case in test_cases:
        assert parser.pypthon_cmd_to_python(*case[:-1], False) == case[-1]
